# cython: language_level=3
import cython
from libc.stdint cimport uint8_t, uint32_t, uint64_t


# cython's way to declare constants
# see https://cython.readthedocs.io/en/latest/src/userguide/language_basics.html#c-variable-and-type-definitions
cdef enum:
    OBJECT_HEADER_SIZE = 12
    EXTENSION_SIZE = 4
    SYNC_MARKER_SIZE = 4

    TYPE_N_BITS = 16
    TYPE_POS = 0

    USER_N_BITS = 1
    USER_POS = 16

    EXTENDED_N_BITS = 1
    EXTENDED_POS = 17

    VERSION_N_BITS = 12
    VERSION_POS = 20

    ONLY_SUBOBJECTS_N_BITS = 1
    ONLY_SUBOBJECTS_POS = 30

    LENGTH_N_BITS = 30
    LENGTH_POS = 0

    EXTENSION_N_BITS = 12
    EXTENSION_POS = 0



cdef class ObjectHeader:
    cdef readonly uint32_t id
    cdef readonly uint32_t type
    cdef readonly uint32_t version
    cdef readonly bint user
    cdef readonly bint extended
    cdef readonly bint only_subobjects
    cdef readonly uint8_t header_size
    cdef public uint64_t content_address
    cdef public uint64_t content_size

    @property
    def total_size(self):
        return self.content_size + self.header_size

    def __repr__(self):
        return (
            'Header[{}]('.format(self.type)
            + 'version={}, '.format(self.version)
            + 'id={}, '.format(self.id)
            + 'user={}, '.format(self.user)
            + 'extended={}, '.format(self.extended)
            + 'only_subobjects={}, '.format(self.only_subobjects)
            + 'header_size{}, '.format(self.header_size)
            + 'content_size={}, '.format(self.content_size)
            + 'content_address={})'.format(self.content_address)
        )


cpdef bint bool_bit_from_pos(uint32_t uint32_word, uint32_t pos):
    '''parse a Python Boolean from a bit a position `pos` in an
    unsigned 32bit integer.
    '''
    return uint32_word & (1 << pos)


cpdef uint32_t get_bits_from_word(
    uint32_t uint32_word,
    uint32_t n_bits,
    uint32_t first,
):
    '''return `n_bits` bits from the input word
    starting at first

    assume the input word was:
        MSB                                    LSB
        0000_0000__0000_0000__1010_1100__0000_0000

    and first=10 and n_bits=4

    the return value would be: 1011 (with leading zeros)
    '''
    return (uint32_word >> first) & ((1 << n_bits) - 1)


cdef (uint32_t, uint32_t, bint, bint) parse_type_field(uint32_t word):
    '''parse TypeInfo
    '''
    type_ = get_bits_from_word(word, TYPE_N_BITS, TYPE_POS)
    user_bit = bool_bit_from_pos(word, USER_POS)
    extended = bool_bit_from_pos(word, EXTENDED_POS)
    version = get_bits_from_word(word, VERSION_N_BITS, VERSION_POS)
    return type_, version, user_bit, extended


@cython.boundscheck(False)
@cython.wraparound(False)
cdef uint32_t unpack_uint32(const uint8_t[:] data):
    return (<uint32_t*> &data[0])[0]


cpdef ObjectHeader parse_header_bytes(const uint8_t[:] header_bytes, bint toplevel=0):
    cdef uint64_t type_int
    cdef uint64_t id_field
    cdef uint64_t length_field

    cdef uint64_t type_
    cdef bint user
    cdef bint extended
    cdef uint64_t version
    cdef bint only_subobjects
    cdef uint64_t length


    type_int = unpack_uint32(header_bytes[0:4])
    type_, version, user, extended = parse_type_field(type_int)
    id_field = unpack_uint32(header_bytes[4:8])

    length_field = unpack_uint32(header_bytes[8:12])

    only_subobjects = bool_bit_from_pos(length_field, ONLY_SUBOBJECTS_POS)
    length = get_bits_from_word(length_field, LENGTH_N_BITS, LENGTH_POS)

    header = ObjectHeader()
    header.type = type_
    header.user = user
    header.extended = extended
    header.version = version
    header.id = id_field
    header.only_subobjects = only_subobjects
    header.content_size = length
    header.header_size = OBJECT_HEADER_SIZE

    if toplevel:
        header.header_size += SYNC_MARKER_SIZE

    if header.extended:
        header.header_size += EXTENSION_SIZE

    return header



import logging
from collections import namedtuple

from ..base import KNOWN_OBJECTS, EventIOFile, EventIOObject
from ..exceptions import check_type

from .objects import (
    RunHeader,
    TelescopeDefinition,
    EventHeader,
    ArrayOffsets,
    TelescopeData,
    Photons,
    CameraLayout,
    TriggerTime,
    PhotoElectrons,
    EventEnd,
    RunEnd,
    Longitudinal,
    InputCard,
    AtmosphericProfile,
)


for o in EventIOObject.__subclasses__():
    KNOWN_OBJECTS[o.eventio_type] = o

log = logging.getLogger(__name__)


__all__ = [
    'IACTFile',
    'RunHeader',
    'TelescopeDefinition',
    'EventHeader',
    'ArrayOffsets',
    'TelescopeData',
    'Photons',
    'CameraLayout',
    'TriggerTime',
    'PhotoElectrons',
    'EventEnd',
    'RunEnd',
    'Longitudinal',
    'InputCard',
    'AtmosphericProfile',
]


class IACTFile(EventIOFile):
    '''
    An Interface to access the data of a EventIO file
    as produced by the CORSIKA IACT (a.k.a. bernlohr) extension
    more easily.

    Instead of low-level access to eventio items, it provides
    direct access to telescope events and simulation settings.

    It is an Iterable of `Event`s.

    Notes
    -----
    Calling `next` on this file will give you the next low-level EventIOObject.
    Calling `next(iter(File))` will give you the next event.

    The structure of an  EventIO file is assumed to be like this:

    RunHeader
    InputCard
    TelescopeDefinition

    For each Event:
      EventHeader
      ArrayOffsets
      Longitudinal (optional)
      For each reuse:
        TelescopeData
        For each Telescope:
          Photons
      EventEnd

    RunEnd
    '''

    def __init__(self, path, zcat=True):
        super().__init__(path, zcat=zcat)

        header_object = next(self)
        check_type(header_object, RunHeader)
        self.header = header_object.parse()

        input_card_object = next(self)
        check_type(input_card_object, InputCard)
        self.input_card = input_card_object.parse()

        o = next(self)
        if isinstance(o, AtmosphericProfile):
            self.atmospheric_profile = o.parse()
            telescope_object = next(self)
        else:
            self.atmospheric_profile = None
            telescope_object = o

        check_type(telescope_object, TelescopeDefinition)
        self.n_telescopes = telescope_object.n_telescopes
        self.telescope_positions = telescope_object.parse()
        self._first_event_byte = self.tell()

    def __repr__(self):
        return (
            '{}(\n'
            '  path={}\n'
            '  n_telescopes={}\n'
            ')'
        ).format(
            self.__class__.__name__,
            self.path,
            self.n_telescopes,
        )

    def __iter__(self):
        '''
        Generator over the single array events
        '''
        self._next_header_pos = self._first_event_byte
        obj = next(self)

        while not isinstance(obj, RunEnd):
            check_type(obj, EventHeader)
            header = obj.parse()

            reuse_object = next(self)
            check_type(reuse_object, ArrayOffsets)

            time_offset, array_offsets = reuse_object.parse()
            n_reuses = len(array_offsets)

            longitudinal = None
            particles = None

            obj = next(self)

            while not isinstance(obj, TelescopeData):
                if isinstance(obj, Longitudinal):
                    longitudinal = obj.parse()

                if isinstance(obj, Photons):
                    if obj.array_id != 999 or obj.telescope_id != 999:
                        raise ValueError('Unexpected Photon Block')

                    particles = obj.parse()

                obj = next(self)

            for reuse in range(n_reuses):

                check_type(obj, TelescopeData)
                telescope_data_obj = obj

                photon_bunches = {}
                emitter_bunches = {}
                n_photons = {}
                n_bunches = {}
                for data in telescope_data_obj:
                    if isinstance(data, Photons):
                        photons, emitter = data.parse()
                        photon_bunches[data.telescope] = photons
                        emitter_bunches[data.telescope] = emitter
                        n_photons[data.telescope] = data.n_photons
                        n_bunches[data.telescope] = data.n_bunches

                if len(array_offsets.dtype) == 3:
                    weight = array_offsets[reuse]['weight']
                else:
                    weight = 1.0

                yield Event(
                    header=header,
                    photon_bunches=photon_bunches,
                    time_offset=time_offset,
                    impact_x=-array_offsets[reuse]['x'],
                    impact_y=-array_offsets[reuse]['y'],
                    reuse_weight=weight,
                    event_number=header['event_number'],
                    reuse=reuse + 1,
                    n_photons=n_photons,
                    n_bunches=n_bunches,
                    longitudinal=longitudinal,
                    particles=particles,
                    emitter=emitter_bunches,
                )

                obj = next(self)

            check_type(obj, EventEnd)

            obj = next(self)

        self.run_end = obj.parse()


EventTuple = namedtuple(
    'EventTuple',
    [
        'header', 'photon_bunches',
        'time_offset', 'impact_x', 'impact_y',
        'reuse_weight',
        'event_number', 'reuse',
        'n_photons', 'n_bunches',
        'longitudinal', 'particles',
        'emitter',
    ]
)


class Event(EventTuple):
    '''
    A single event as simulated by corsika

    Members:
      event_number

      shower:
          the id of the simulated shower

      reuse:
          reuse index for this shower

      header:
        a dictionary containing the corsika event header

      end_block:
        numpy array of floats with the event end block data

      photon_bunches:
        a dictionary mapping telescope_ids to numpy
        arrays with the photon bunch data with the following colums:
          x:         x coordinate in cm
          y:         y coordinate in cm
          cx:        cosine of incident angle in x direction
          cy:        cosine of incident angle in y direction
          time:      time since first interaction in ns
          zem:       Emission height in cm above sea level
          lambda:    wavelength in nm
          scattered: indicates if the photon was scattered in the atmosphere

      time_offset:
        time from first interaction to ground in ns

      x_offset:
        array offset in x direction in cm

      y_offset:
        array offset in y direction in cm

      weight:
        weight for this offset position.
        Only different from 1 if importance sampling was used.
    '''
    def __repr__(self):
        return '{}(event_number={}, reuse={}, n_telescopes={}, n_photons={})'.format(
            self.__class__.__name__,
            self.event_number,
            self.reuse,
            len(self.n_bunches),
            self.n_photons,
        )



# cython: language_level=3
import cython
from cpython cimport array
from libc.stdint cimport uint8_t, uint32_t, uint64_t, int16_t, int32_t
import numpy as np
cimport numpy as np

INT16 = np.int16

@cython.wraparound(False)  # disable negative indexing
cpdef read_sector_information_v1(
    const uint8_t[:] data,
    uint64_t n_pixels,
    uint64_t offset = 0,
):
    cdef uint64_t i
    cdef int32_t j, n
    cdef int16_t* short_ptr

    cdef list sectors = []
    cdef array.array sector

    cdef uint64_t pos = offset
    for i in range(n_pixels):

        short_ptr = <int16_t*> &data[pos]
        n = short_ptr[0]
        pos += 2

        sector = array.array('h')
        for j in range(n):
            short_ptr = <int16_t*> &data[pos]
            sector.append(short_ptr[0])
            pos += 2

        # FIXME:
        # according to a comment in the c-sources
        # there is might be an old bug here,
        # which is trailing zeros.
        # is an ascending list of numbes, so any zero
        # after the first position indicates the end of sector.
        #
        # DN: maybe this bug was fixed long ago,
        # so maybe we do not have to account for it here
        # I will check for it in the tests.
        sectors.append(sector)

    return sectors, pos - offset



@cython.wraparound(False)  # disable negative indexing
cpdef dict parse_mc_event(
    const uint8_t[:] data,
    uint32_t version
):

    cdef uint64_t pos = 0
    cdef float xcore, ycore, aweight
    cdef int32_t shower_num

    shower_num = (<int32_t*> &data[pos])[0]
    pos += 4
    xcore = (<float*> &data[pos])[0]
    pos += 4
    ycore = (<float*> &data[pos])[0]
    pos += 4

    if version >= 2:
        aweight = (<float*> &data[pos])[0]
    else:
        aweight = 1.0

    return {
        'shower_num': shower_num,
        'xcore': xcore,
        'ycore': ycore,
        'aweight': aweight
    }



# cython: language_level=3
import cython
import numpy as np
cimport numpy as np
from libc.stdint cimport uint8_t, uint16_t, uint32_t, uint64_t, int16_t, int32_t, int64_t

cdef INT16 = np.int16
cdef UINT32 = np.uint32
cdef INT32 = np.int32
cdef UINT64 = np.uint64
cdef INT64 = np.int64
cdef F32 = np.float32


@cython.wraparound(False)  # disable negative indexing
cpdef (uint64_t, uint8_t) unsigned_varint(const uint8_t[:] data, uint64_t offset=0):
    cdef uint8_t length
    cdef uint64_t value
    length = get_length_of_varint(data[offset])
    value = parse_varint(data[offset:offset + length])
    return value, length


@cython.wraparound(False)  # disable negative indexing
cpdef (int64_t, uint8_t) varint(const uint8_t[:] data, uint64_t offset=0):
    cdef uint8_t length
    cdef uint64_t value
    value, length = unsigned_varint(data, offset)
    if (value & 1) == 1:  # Negative number;
        return -(value >> 1) - 1, length
    return value >> 1, length


cpdef uint8_t get_length_of_varint(const uint8_t first_byte):
    if (first_byte & 0x80) == 0:
        return 1
    if (first_byte & 0xc0) == 0x80:
        return 2
    if (first_byte & 0xe0) == 0xc0:
        return 3
    if (first_byte & 0xf0) == 0xe0:
        return 4
    if (first_byte & 0xf8) == 0xf0:
        return 5
    if (first_byte & 0xfc) == 0xf8:
        return 6
    if (first_byte & 0xfe) == 0xfc:
        return 7
    if (first_byte & 0xff) == 0xfe:
        return 8
    return 9


@cython.boundscheck(False)
@cython.wraparound(False)  # disable negative indexing
cpdef uint64_t parse_varint(const uint8_t[:] var_int_bytes):
    cdef uint8_t length = var_int_bytes.shape[0]
    cdef uint64_t v[9]
    cdef uint8_t i  = 0
    for i in range(length):
        v[i] = var_int_bytes[i]

    if length == 1:
        return v[0]

    if length == 2:
        return ((v[0] & 0x3f) <<8) | v[1]

    if length == 3:
        return (
            ((v[0] & 0x1f) << 16)
            | (v[1] << 8)
            | v[2]
        )

    if length == 4:
        return (
            ((v[0] & 0x0f) << 24)
            | (v[1] << 16)
            | (v[2] << 8)
            | v[3]
        )
    if length == 5:
        return (
            ((v[0] & 0x07) << 32)
            | (v[1] << 24)
            | (v[2] << 16)
            | (v[3] << 8)
            | v[4]
        )
    if length == 6:
        return (
            ((v[0] & 0x03) << 40)
            | (v[1] << 32)
            | (v[2] << 24)
            | (v[3] << 16)
            | (v[4] << 8)
            | v[5]
        )
    if length == 7:
        return (
            ((v[0] & 0x01) << 48)
            | (v[1] << 40)
            | (v[2] << 32)
            | (v[3] << 24)
            | (v[4] << 16)
            | (v[5] << 8)
            | v[6]
        )
    if length == 8:
        return (
            (v[1]<<48)
            | (v[2]<<40)
            | (v[3]<<32)
            | (v[4]<<24)
            | (v[5]<<16)
            | (v[6]<<8)
            | v[7]
        )

    return (
        (v[1]<<56)
        | (v[2]<<48)
        | (v[3]<<40)
        | (v[4]<<32)
        | (v[5]<<24)
        | (v[6]<<16)
        | (v[7]<<8)
        | v[8]
    )


@cython.wraparound(False)  # disable negative indexing
cpdef unsigned_varint_array(
    const uint8_t[:] data,
    uint64_t n_elements,
    uint64_t offset = 0,
):
    cdef np.ndarray[uint64_t, ndim=1] output = np.empty(n_elements, dtype=UINT64)

    cdef uint64_t i
    cdef uint64_t pos
    cdef uint64_t idx
    cdef uint64_t length
    pos = 0

    for i in range(n_elements):
        idx = pos + offset
        length = get_length_of_varint(data[idx])
        output[i] = parse_varint(data[idx:idx + length])
        pos += length

    return output, pos


@cython.wraparound(False)  # disable negative indexing
cpdef varint_array(
    const uint8_t[:] data,
    uint64_t n_elements,
    uint64_t offset = 0,
):
    cdef uint64_t bytes_read
    cdef np.ndarray[int64_t, ndim=1] output = np.empty(n_elements, dtype=INT64)

    cdef int64_t val
    cdef uint8_t length

    cdef uint8_t one = 1;
    cdef uint64_t i
    cdef uint64_t pos = offset

    for i in range(n_elements):
        length = get_length_of_varint(data[pos])
        val = parse_varint(data[pos:pos + length])
        pos += length

        if (val & one):
            output[i] = -(val >> one) - one
        else:
            output[i] = val >> one

    return output, pos - offset


@cython.wraparound(False)  # disable negative indexing
cpdef unsigned_varint_arrays_differential(
    const uint8_t[:] data,
    uint64_t n_arrays,
    uint64_t n_elements,
    uint64_t offset = 0,
):
    cdef uint64_t pos = 0
    cdef uint64_t bytes_read = 0
    cdef uint64_t bytes_read_total = 0
    cdef uint64_t i
    cdef uint64_t j
    cdef (uint64_t, uint64_t) shape = (n_arrays, n_elements)

    cdef np.ndarray[uint32_t, ndim=2] output = np.zeros(shape, dtype=UINT32)
    cdef uint32_t[:, :] output_view = output
    cdef uint32_t[:] output_view_1d

    for i in range(n_arrays):

        output_view_1d = output_view[i]

        bytes_read = unsigned_varint_array_differential(
            data, output=output_view_1d, offset=offset
        )
        offset += bytes_read
        bytes_read_total += bytes_read

    return output, bytes_read_total


@cython.wraparound(False)  # disable negative indexing
cdef uint64_t unsigned_varint_array_differential(
    const uint8_t[:] data,
    uint32_t[:] output,
    uint64_t offset = 0,
) except -1:

    cdef uint64_t n_elements = output.shape[0]
    cdef int32_t val = 0
    cdef uint64_t i
    cdef uint64_t pos = 0
    cdef uint8_t v0, v1, v2, v3, v4

    for i in range(n_elements):
        v0 = data[pos + offset]
        pos += 1

        if (v0 & 0x80) == 0:  # one byte
            if (v0 & 0x01) == 0:  # positive
                val += v0 >> 1
            else:  # negative
                val -= (v0 >> 1) + 1
        elif (v0 & 0xc0) == 0x80:  # two bytes
            v1 = data[pos + offset]
            pos += 1
            if (v1 & 0x01) == 0:  # positive
                val += ((v0 & 0x3f) << 7) | (v1 >> 1)
            else:  # negative
                val -= (((v0 & 0x3f) << 7) | (v1 >> 1)) + 1
        elif (v0 & 0xe0) == 0xc0:  # three bytes
            v1 = data[pos + offset + 0]
            v2 = data[pos + offset + 1]
            pos += 2

            if (v2 & 0x01) == 0:
                val += (
                    ((v0 & 0x1f) << 15)
                    | (v1 << 7)
                    | (v2 >> 1)
                )
            else:
                val -= (
                    ((v0 & 0x1f) << 15)
                    | (v1 << 7)
                    | (v2 >> 1)
                ) + 1
        elif (v0 & 0xf0) == 0xe0:  # four bytes
            v1 = data[pos + offset + 0]
            v2 = data[pos + offset + 1]
            v3 = data[pos + offset + 2]
            pos += 3
            if (v3 & 0x01) == 0:
                val += (
                    ((v0 & 0x0f) << 23)
                    | (v1 << 15)
                    | (v2 << 7)
                    | (v3 >> 1)
                )
            else:
                val -= (
                    ((v0 & 0x0f) << 23)
                    | (v1 << 15)
                    | (v2 << 7)
                    | (v3 >> 1)
                ) + 1
        elif (v0 & 0xf8) == 0xf0:
            v1 = data[pos + offset + 0]
            v2 = data[pos + offset + 1]
            v3 = data[pos + offset + 2]
            v4 = data[pos + offset + 3]
            pos += 4
            # The format would allow bits 32 and 33 being set but we ignore this here. */
            if (v4 & 0x01) == 0:
                val += (
                    ((v0 & 0x07) << 31)
                    | (v1 << 23)
                    | (v2 << 15)
                    | (v3 << 7)
                    | (v4 >> 1)
                )
            else:
                val -= (
                    ((v0 & 0x07) << 31)
                    | (v1 << 23)
                    | (v2 << 15)
                    | (v3 << 7)
                    | (v4 >> 1)
                ) + 1
        output[i] = val

    return pos



def simtel_pixel_timing_parse_list_type_2(
    const uint8_t[:] data,
    const int16_t[:, :] pixel_list,
    uint32_t n_gains,
    uint32_t n_pixels,
    uint32_t n_types,
    bint with_sum,
    bint glob_only_selected,
    float granularity,
):
    cdef int16_t start, stop
    cdef uint32_t list_index
    cdef uint32_t n_lists = pixel_list.shape[0]
    cdef uint32_t i_pix, i_type
    cdef uint64_t pos = 0
    cdef uint32_t length = 0

    cdef np.ndarray[float, ndim=2] timval = np.full((n_pixels, n_types), np.nan, dtype=np.float32)
    cdef np.ndarray[int32_t, ndim=2] pulse_sum_loc = np.zeros((n_gains, n_pixels), dtype=INT32)
    cdef np.ndarray[int32_t, ndim=2] pulse_sum_glob = np.zeros((n_gains, n_pixels), dtype=INT32)

    for list_index in range(n_lists):
        start = pixel_list[list_index][0]
        stop = pixel_list[list_index][1]

        for i_pix in range(start, stop + 1):
            for i_type in range(n_types):
                timval[i_pix, i_type] = granularity * (<int16_t*> &data[pos])[0]
                pos += 2

            if with_sum:
                for i_gain in range(n_gains):
                    pulse_sum_loc[i_gain, i_pix], length = varint(data, offset=pos)
                    pos += length

                if glob_only_selected:
                    for i_gain in range(n_gains):
                        pulse_sum_glob[i_gain, i_pix], length = varint(data, offset=pos)
                        pos += length

    if with_sum and pixel_list.shape[0] > 0 and not glob_only_selected:
        for i_gain in range(n_gains):
            for i_pix in range(n_pixels):
                pulse_sum_glob[i_gain, i_pix], length = varint(data, offset=pos)
                pos += length

    return {
        'time': timval,
        'pulse_sum_glob': pulse_sum_glob,
        'pulse_sum_loc': pulse_sum_loc,
    }, pos


def parse_1208(
    const uint8_t[:] data,
    uint32_t n_pixels,
    uint32_t nonempty,
    uint32_t version,
    uint32_t flags
):

    cdef uint32_t length

    cdef np.ndarray[float, ndim=1] photoelectrons = np.zeros(n_pixels, dtype=F32)
    cdef np.ndarray[int32_t, ndim=1] photon_counts = None
    cdef list time = [[] for _ in range(n_pixels)]
    cdef list amplitude

    if flags & 1:
        amplitude = [[] for _ in range(n_pixels)]
    else:
        amplitude = None

    cdef uint64_t pos = 0
    cdef uint32_t i
    cdef int32_t j

    for i in range(nonempty):
        if version > 2:
            pix_id, length = varint(data, offset=pos)
        else:
            pix_id = (<int16_t*> &data[pos])[0]
            length = 2

        pos += length

        n_pe = (<int32_t*> &data[pos])[0]
        pos += 4

        photoelectrons[pix_id] = n_pe
        for j in range(n_pe):
            time[pix_id].append((<float*> &data[pos])[0])
            pos += 4

        if flags & 1:
            for j in range(n_pe):
                amplitude[pix_id].append((<float*> &data[pos])[0])
                pos += 4

    if flags & 4:
        photon_counts = np.zeros(n_pixels, dtype=INT32)

        int_ptr = <int32_t*> &data[pos]
        nonempty = int_ptr[0]
        pos += 4

        for i in range(nonempty):
            pix_id = (<int16_t*> &data[pos])[0]
            pos += 2

            photon_counts[pix_id] = (<int32_t*> &data[pos])[0]
            pos += 4

    return photoelectrons, time, amplitude, photon_counts


cpdef simtel_pixel_timing_parse_list_type_1(
    const uint8_t[:] data,
    const int16_t[:] pixel_list,
    uint32_t n_gains,
    uint32_t n_pixels,
    uint32_t n_types,
    bint with_sum,
    bint glob_only_selected,
    float granularity,
):
    cdef uint32_t start, stop, list_index
    cdef uint32_t pixel_list_length = pixel_list.shape[0]
    cdef uint32_t i, i_gain, i_type, i_pix
    cdef uint32_t length = 0
    cdef int16_t* short_ptr

    cdef np.ndarray[float, ndim=2] timval = np.full((n_pixels, n_types), np.nan, dtype=F32)
    cdef np.ndarray[int32_t, ndim=2] pulse_sum_loc = np.zeros((n_gains, n_pixels), dtype=INT32)
    cdef np.ndarray[int32_t, ndim=2] pulse_sum_glob = np.zeros((n_gains, n_pixels), dtype=INT32)


    cdef uint64_t pos = 0
    for i in range(pixel_list_length):
        i_pix = pixel_list[i]
        for i_type in range(n_types):
            short_ptr = <int16_t*> &(data[pos])
            timval[i_pix, i_type] = granularity * short_ptr[0]
            pos += 2

        if with_sum:
            for i_gain in range(n_gains):
                pulse_sum_loc[i_gain, i_pix], length = varint(data, offset=pos)
                pos += length

            if glob_only_selected:
                for i_gain in range(n_gains):
                    pulse_sum_glob[i_gain, i_pix], length = varint(data, offset=pos)
                    pos += length

    if with_sum and len(pixel_list) > 0 and not glob_only_selected:
        for i_gain in range(n_gains):
            for i_pix in range(n_pixels):
                pulse_sum_glob[i_gain, i_pix], length = varint(
                    data, offset=pos,
                )
                pos += length

    return {
        'time': timval,
        'pulse_sum_glob': pulse_sum_glob,
        'pulse_sum_loc': pulse_sum_loc,
    }, pos



@cython.wraparound(False)  # disable negative indexing
cpdef read_sector_information_v2(
    const uint8_t[:] data,
    uint32_t n_pixels,
    uint64_t offset = 0,
):
    cdef uint32_t i
    cdef uint64_t length
    cdef int64_t n

    cdef list sectors = []
    cdef np.ndarray[int64_t, ndim=1] sector

    cdef uint64_t pos = offset
    for i in range(n_pixels):

        n, length = varint(data, offset=pos)
        pos += length

        sector, length = varint_array(data, n, offset=pos)
        pos += length
        sectors.append(sector)

    return sectors, pos - offset



from eventio import EventIOFile
import warnings


def find_type(f, eventio_type):
    o = next(f)
    while not isinstance(o, eventio_type):
        o = next(f)

    if not isinstance(o, eventio_type):
        raise ValueError('Type {} not found'.format(eventio_type))

    return o


def collect_toplevel_of_type(f, eventio_type):
    return list(yield_toplevel_of_type(f, eventio_type))


def yield_toplevel_of_type(f, eventio_type):
    try:
        for o in f:
            if isinstance(o, eventio_type):
                yield o
    except EOFError:
        warnings.warn("File seems to be truncated")


def find_all_subobjects(f, structure, level=0):
    '''
    Find all subobjects expected in structure.
    So if you want all AdcSums, use
    structure = [ArrayEvent, TelescopeEvent, ADCSums]
    '''
    objects = []
    elem = structure[level]

    try:
        for o in f:
            if isinstance(o, structure[-1]):
                objects.append(o)
            elif isinstance(o, elem):
                objects.extend(find_all_subobjects(o, structure, level + 1))
    except EOFError:
        warnings.warn("File seems to be truncated")
    return objects


def yield_all_subobjects(f, structure, level=0):
    '''
    Find all subobjects expected in structure.
    So if you want all AdcSums, use
    structure = [ArrayEvent, TelescopeEvent, ADCSums]
    '''
    elem = structure[level]

    for o in f:
        if isinstance(o, structure[-1]):
            yield o
        elif isinstance(o, elem):
            yield from yield_all_subobjects(o, structure, level + 1)


def yield_all_objects_depth_first(f, level=0):
    '''yield subobjects of type, regardless of structure'''

    if isinstance(f, EventIOFile) or f.header.only_subobjects:
        for o in f:
            yield o, level
            yield from yield_all_objects_depth_first(o, level + 1)


def yield_subobjects(f, eventio_type):
    '''yield subobjects of type, regardless of structure'''
    if isinstance(f, eventio_type):
        yield f
    else:
        if isinstance(f, EventIOFile) or f.header.only_subobjects:
            for o in f:
                yield from yield_subobjects(o, eventio_type)


def yield_n_subobjects(f, eventio_type, n=3):
    for i, obj in enumerate(yield_subobjects(f, eventio_type), start=1):
        if i >= n:
            yield obj
            return

        yield obj



'''
Implementation of an EventIOFile that
loops through SimTel Array events.
'''
import re
from copy import copy
from collections import defaultdict
import logging
from ..base import EventIOFile
from ..exceptions import check_type
from .. import iact
from ..histograms import Histograms
from .objects import (
    ADCSamples,
    ADCSums,
    ArrayEvent,
    CalibrationEvent,
    CameraMonitoring,
    CameraOrganization,
    CameraSettings,
    CameraSoftwareSettings,
    DisabledPixels,
    DriveSettings,
    History,
    ImageParameters,
    LaserCalibration,
    MCEvent,
    MCPhotoelectronSum,
    MCRunHeader,
    MCShower,
    PixelList,
    PixelSettings,
    PixelTiming,
    PixelTriggerTimes,
    PointingCorrection,
    RunHeader,
    TelescopeEvent,
    TelescopeEventHeader,
    TrackingPosition,
    TriggerInformation,
)


telescope_descriptions_types = (
    CameraSettings,
    CameraOrganization,
    PixelSettings,
    DisabledPixels,
    CameraSoftwareSettings,
    DriveSettings,
    PointingCorrection,
)


log = logging.getLogger(__name__)


camel_re1 = re.compile('(.)([A-Z][a-z]+)')
camel_re2 = re.compile('([a-z0-9])([A-Z])')


def camel_to_snake(name):
    s1 = camel_re1.sub(r'\1_\2', name)
    return camel_re2.sub(r'\1_\2', s1).lower()


class NoTrackingPositions(Exception):
    pass


class SimTelFile(EventIOFile):
    def __init__(self, path, allowed_telescopes=None, skip_calibration=False, zcat=True):
        super().__init__(path, zcat=zcat)

        self.path = path
        self.allowed_telescopes = None
        if allowed_telescopes:
            self.allowed_telescopes = set(allowed_telescopes)

        self.histograms = None

        self.history = []
        self.mc_run_headers = []
        self.corsika_inputcards = []
        self.atmospheric_profiles = []
        self.header = None
        self.n_telescopes = None
        self.telescope_descriptions = defaultdict(dict)
        self.camera_monitorings = defaultdict(dict)
        self.laser_calibrations = defaultdict(dict)
        self.current_mc_shower = None
        self.current_mc_event = None
        self.current_photoelectron_sum = None
        self.current_photoelectrons = {}
        self.current_array_event = None
        self.current_calibration_event = None
        self.skip_calibration = skip_calibration

        # read the header:
        # assumption: the header is done when
        # any of the objects in check is not None anymore
        # and we found the telescope_descriptions of all telescopes
        check = []
        found_all_telescopes = False
        while not (any(o for o in check) and found_all_telescopes):
            self.next_low_level()

            check = [
                self.current_mc_shower,
                self.current_array_event,
                self.current_calibration_event,
                self.laser_calibrations,
                self.camera_monitorings,
            ]

            # check if we found all the descriptions of all telescopes
            if self.n_telescopes is not None:
                found = sum(
                    len(t) == len(telescope_descriptions_types)
                    for t in self.telescope_descriptions.values()
                )
                found_all_telescopes = found == self.n_telescopes

    def __iter__(self):
        return self.iter_array_events()

    def next_low_level(self):
        o = next(self)

        # order of if statements is roughly sorted
        # by the number of occurences in a simtel file
        # this should minimize the number of if statements evaluated

        if isinstance(o, MCEvent):
            self.current_mc_event = o.parse()
            self.current_mc_event_id = o.header.id

        elif isinstance(o, MCShower):
            self.current_mc_shower = o.parse()

        elif isinstance(o, ArrayEvent):
            self.current_array_event = parse_array_event(
                o,
                self.allowed_telescopes
            )

        elif isinstance(o, iact.TelescopeData):
            self.current_photoelectrons = parse_photoelectrons(o)

        elif isinstance(o, CameraMonitoring):
            self.camera_monitorings[o.telescope_id].update(o.parse())

        elif isinstance(o, LaserCalibration):
            self.laser_calibrations[o.telescope_id].update(o.parse())

        elif isinstance(o, telescope_descriptions_types):
            key = camel_to_snake(o.__class__.__name__)
            self.telescope_descriptions[o.telescope_id][key] = o.parse()

        elif isinstance(o, RunHeader):
            self.header = o.parse()
            self.n_telescopes = self.header['n_telescopes']

        elif isinstance(o, MCRunHeader):
            self.mc_run_headers.append(o.parse())

        elif isinstance(o, iact.InputCard):
            self.corsika_inputcards.append(o.parse())
        elif isinstance(o, MCPhotoelectronSum):
            self.current_photoelectron_sum = o.parse()

        elif isinstance(o, CalibrationEvent):
            if not self.skip_calibration:
                self.current_calibration_event = parse_array_event(
                    next(o),
                    self.allowed_telescopes,
                )

        elif isinstance(o, History):
            self.history.append(o)

        elif isinstance(o, Histograms):
            self.histograms = o.parse()
        elif isinstance(o, iact.AtmosphericProfile):
            self.atmospheric_profiles.append(o.parse())
        else:
            raise Exception(
                'object type encountered, which is no handled'
                'at the moment: {}'.format(o)
            )

    def iter_mc_events(self):
        while True:
            try:
                next_event = self.try_build_mc_event()
            except StopIteration:
                break
            if next_event is not None:
                yield next_event

    def try_build_mc_event(self):
        if self.current_mc_event:
            event_data = {
                'event_id': self.current_mc_event_id,
                'mc_shower': self.current_mc_shower,
                'mc_event': self.current_mc_event,
            }
            self.current_mc_event = None
            return event_data
        self.next_low_level()

    def iter_array_events(self):
        while True:

            next_event = self.try_build_event()
            if next_event is not None:
                yield next_event

            try:
                self.next_low_level()
            except StopIteration:
                break

    def try_build_event(self):
        '''check if all necessary info for an event was found,
        then make an event and invalidate old data
        '''
        if self.current_array_event:
            if (
                self.allowed_telescopes
                and not self.current_array_event['telescope_events']
            ):
                self.current_array_event = None
                return None

            event_data = {
                'type': 'data',
                'event_id': self.current_mc_event_id,
                'mc_shower': self.current_mc_shower,
                'mc_event': self.current_mc_event,
                'telescope_events': self.current_array_event['telescope_events'],
                'tracking_positions': self.current_array_event['tracking_positions'],
                'trigger_information': self.current_array_event['trigger_information'],
                'photoelectron_sums': self.current_photoelectron_sum,
                'photoelectrons': self.current_photoelectrons,
            }

            event_data['camera_monitorings'] = {
                telescope_id: copy(self.camera_monitorings[telescope_id])
                for telescope_id in self.current_array_event['telescope_events'].keys()
            }
            event_data['laser_calibrations'] = {
                telescope_id: copy(self.laser_calibrations[telescope_id])
                for telescope_id in self.current_array_event['telescope_events'].keys()
            }

            self.current_array_event = None

            return event_data

        elif self.current_calibration_event:
            event = self.current_calibration_event
            if (
                self.allowed_telescopes
                and not self.current_array_event['telescope_events']
            ):
                self.current_calibration_event = None
                return None

            event_data = {
                'type': 'calibration',
                'telescope_events': event['telescope_events'],
                'tracking_positions': event['tracking_positions'],
                'trigger_information': event['trigger_information'],
            }

            event_data['camera_monitorings'] = {
                telescope_id: copy(self.camera_monitorings[telescope_id])
                for telescope_id in event['telescope_events'].keys()
            }
            event_data['laser_calibrations'] = {
                telescope_id: copy(self.laser_calibrations[telescope_id])
                for telescope_id in event['telescope_events'].keys()
            }

            self.current_calibration_event = None

            return event_data


def parse_array_event(array_event, allowed_telescopes=None):
    '''structure of event:
        TriggerInformation[2009]  <-- this knows how many TelescopeEvents

        TelescopeEvent[2202]
        ...
        TelescopeEvent[2208]

        TrackingPosition[2101]
        ...
        TrackingPosition[2164]

        StereoReconstruction[2015]


        In words:
            1 cent event
            n tel events
            m track events (n does not need to be == m)
            1 shower
    '''
    check_type(array_event, ArrayEvent)

    telescope_events = {}
    tracking_positions = {}

    for i, o in enumerate(array_event):
        # require first element to be a TriggerInformation
        if i == 0:
            check_type(o, TriggerInformation)
            trigger_information = o.parse()
            telescopes = set(trigger_information['telescopes_with_data'])

            if allowed_telescopes and len(telescopes & allowed_telescopes) == 0:
                break

        elif isinstance(o, TelescopeEvent):
            if allowed_telescopes is None or o.telescope_id in allowed_telescopes:
                telescope_events[o.telescope_id] = parse_telescope_event(o)

        elif isinstance(o, TrackingPosition):
            if allowed_telescopes is None or o.telescope_id in allowed_telescopes:
                tracking_positions[o.telescope_id] = o.parse()

    missing_tracking = set(telescope_events.keys()) - set(tracking_positions.keys())
    if missing_tracking:
        raise NoTrackingPositions(
            'Missing tracking positions for telescopes {}'.format(
                missing_tracking
            )
        )

    return {
        'trigger_information': trigger_information,
        'telescope_events': telescope_events,
        'tracking_positions': tracking_positions,
    }


def parse_photoelectrons(telescope_data):
    check_type(telescope_data, iact.TelescopeData)

    photo_electrons = {}
    for o in telescope_data:
        check_type(o, iact.PhotoElectrons)
        photo_electrons[o.telescope_id] = o.parse()

    return photo_electrons


def parse_telescope_event(telescope_event):
    '''Parse a telescope event'''
    check_type(telescope_event, TelescopeEvent)

    event = {'pixel_lists': {}}
    for i, o in enumerate(telescope_event):

        if i == 0:
            check_type(o, TelescopeEventHeader)
            event['header'] = o.parse()

        elif isinstance(o, ADCSamples):
            event['adc_samples'] = o.parse()

        elif isinstance(o, ADCSums):
            event['adc_sums'] = o.parse()

        elif isinstance(o, PixelTiming):
            event['pixel_timing'] = o.parse()

        elif isinstance(o, ImageParameters):
            event['image_parameters'] = o.parse()

        elif isinstance(o, PixelList):
            event['pixel_lists'][o.code] = o.parse()

        elif isinstance(o, PixelTriggerTimes):
            event['pixel_trigger_times'] = o.parse()

    return event



import struct
import gzip
import logging
import subprocess as sp
import warnings

from .file_types import is_gzip, is_eventio, is_zstd
from .header import parse_header_bytes, get_bits_from_word
from . import constants
from .exceptions import WrongType

try:
    import zstandard as zstd
    has_zstd = True
    version = tuple(int(v) for v in zstd.__version__.split('.'))
    if version < (0, 11, 1):
        warnings.warn('zstandard < 0.11.1 has a memory leaking bug, please update')
except ImportError:
    has_zstd = False

log = logging.getLogger(__name__)

KNOWN_OBJECTS = {}


class PipeWrapper:
    '''
    This class makes a sp.PIPE  forward-seekable
    by keeping track of the bytes already read
    and reading as many bytes as required in `seek`
    '''
    def __init__(self, pipe):
        self.pipe = pipe
        self.pos = 0

    def read(self, size=-1):
        data = self.pipe.read(size)
        self.pos += len(data)
        return data

    def seek(self, offset, whence=0):
        if whence == 0:
            to_read = offset - self.pos
            if to_read < 0:
                raise IOError('Only forward seeking possible')
            self.pos += len(self.pipe.read(to_read))

        if whence == 1:
            if offset < 0:
                raise IOError('Only forward seeking possible')
            self.pos += len(self.pipe.read(offset))

        if whence == 2:
            raise IOError('Only forward seeking possible')

        return self.pos

    def close(self):
        return self.pipe.close()

    def tell(self):
        return self.pos


class EventIOFile:

    def __init__(self, path, zcat=True):
        log.info('Opening new file {}'.format(path))
        self.path = path
        self.read_process = None
        self.zstd = False

        if not is_eventio(path):
            raise ValueError('File {} is not an eventio file'.format(path))

        if is_gzip(path):
            log.info('Found gzipped file')
            if zcat:
                try:
                    log.info('Trying to read using zcat')
                    self.read_process = sp.Popen(
                        ['gzip', '-cd', path], stdout=sp.PIPE, stderr=sp.PIPE
                    )
                    self.read_process.poll()
                    rc = self.read_process.returncode
                    if rc is not None and rc != 0:
                        raise ValueError(self.read_process.stderr.read().decode())

                    self._filehandle = PipeWrapper(self.read_process.stdout)
                    log.info('Using zcat')
                except Exception as e:
                    log.info(str(e))
                    log.warning('Falling back to gzip module')
                    self._filehandle = gzip.open(path)
            else:
                log.info('Using gzip module')
                self._filehandle = gzip.open(path)

        elif is_zstd(path):
            log.info('Found zstd compressed file')
            if not has_zstd:
                raise IOError(
                    'You need to install the `zstandard` module'
                    'to read zstd compressed file'
                )
            self._filehandle = zstd.ZstdDecompressor().stream_reader(open(path, 'rb'))
            self.zstd = True

        else:
            log.info('Found uncompressed file')
            self._filehandle = open(path, mode='rb')

        self._next_header_pos = 0

    def __iter__(self):
        return self

    def __next__(self):
        self.seek(self._next_header_pos)
        read_sync_marker(self)
        header = read_header(
            self,
            toplevel=True,
            offset=self._next_header_pos,
        )
        self._next_header_pos += header.total_size

        return KNOWN_OBJECTS.get(header.type, EventIOObject)(
            header,
            filehandle=self._filehandle,
        )

    def seek(self, position, whence=0):
        return self._filehandle.seek(position, whence)

    def tell(self):
        return self._filehandle.tell()

    def read(self, size=-1):
        return self._filehandle.read(size)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    def close(self):
        self._filehandle.close()
        if self.read_process is not None:
            self.read_process.terminate()


def check_size_or_raise(data, expected_length, zero_ok=True):
    length = len(data)
    if length == 0:
        if zero_ok:
            raise StopIteration
        else:
            raise EOFError('File seems to be truncated')

    if length < expected_length:
        raise EOFError('File seems to be truncated')


def read_sync_marker(byte_stream):
    '''
    Read the sync marker from the filehandle.
    Assumes position of `byte_stream` is at the beginning of a new toplevel header.

    Raises stop iteration if not enough data is available.
    Raises NotImplementedError if BigEndian sync marker is encountered
    '''
    sync = byte_stream.read(constants.SYNC_MARKER_SIZE)
    check_size_or_raise(sync, constants.SYNC_MARKER_SIZE, zero_ok=True)
    check_sync_bytes(sync)


def read_header(byte_stream, offset, toplevel=False):
    '''Read the next header object from the file
    Assumes position of `byte_stream` is at the beginning of a new header.

    Raises stop iteration if not enough data is available.
    '''

    header_bytes = byte_stream.read(constants.OBJECT_HEADER_SIZE)
    check_size_or_raise(
        header_bytes,
        constants.OBJECT_HEADER_SIZE,
        zero_ok=False,
    )

    header = parse_header_bytes(header_bytes, toplevel=toplevel)

    if header.extended:
        extension_field = byte_stream.read(constants.EXTENSION_SIZE)
        check_size_or_raise(
            extension_field,
            constants.EXTENSION_SIZE,
            zero_ok=True
        )
        ext = parse_extension_field(extension_field)
        header.content_size += ext

    header.content_address = offset + header.header_size

    return header


def check_sync_bytes(sync):
    ''' returns the endianness as given by the sync byte '''

    if sync == constants.SYNC_MARKER_LITTLE_ENDIAN:
        return '<'

    if sync == constants.SYNC_MARKER_BIG_ENDIAN:
        raise NotImplementedError(
            'Big endian byte order is not supported by this reader'
        )

    raise ValueError(
        'Sync must be 0xD41F8A37 or 0x378A1FD4. Got: {}'.format(sync)
    )


class EventIOObject:
    '''
    Base Class for eventio objects.
    Can be subclassed to implement different types of
    EventIO objects and how their data payload is parsed into
    python objects.

    EventIO objects can basically play two roles:
        - a binary data blob
        - A list of other `EventIOObject`s

    If an `EventIOObject` is a pure list of other `EventIOObject`s,
    it can be iterated.
    Otherwise, parsing of the binary data has to be implented in a subclass
    or done "by hand" after reading the payload bytes.
    '''
    eventio_type = None

    def __init__(self, header, filehandle):
        if self.eventio_type is not None and header.type != self.eventio_type:
            raise WrongType(self.eventio_type, header.type)

        self._filehandle = filehandle
        self.header = header
        self.address = self.header.content_address
        self.size = self.header.content_size
        self.only_subobjects = self.header.only_subobjects
        self._next_header_pos = 0
        self._pos = 0

    def read(self, size=-1):
        '''Read bytes from the payload of this object.

        Parameters
        ----------
        size: int
            read `size` bytes from the payload of this object
            If size == -1 (default), read all remaining bytes.
        '''
        # read all remaining bytes.
        remaining = self.size - self._pos
        if size == -1 or size > remaining:
            size = remaining

        data = self._filehandle.read(size)
        self._pos += len(data)

        return data

    def __iter__(self):
        if not self.header.only_subobjects:
            raise ValueError(
                'Only EventIOObjects that contain just subobjects are iterable'
            )
        return self

    def __next__(self):
        if not self.only_subobjects:
            raise ValueError(
                'Only EventIOObjects that contain just subobjects are iterable'
            )

        if self._next_header_pos >= self.size:
            raise StopIteration

        self.seek(self._next_header_pos)
        header = read_header(
            self,
            toplevel=False,
            offset=self.address + self._next_header_pos,
        )
        self._next_header_pos += header.total_size

        return KNOWN_OBJECTS.get(header.type, EventIOObject)(
            header, filehandle=self._filehandle
        )

    def seek(self, offset, whence=0):
        address = self.address
        if whence == 0:
            assert offset >= 0
            new_pos = self._filehandle.seek(address + offset, whence)
        elif whence == 1:
            new_pos = self._filehandle.seek(offset, whence)
        elif whence == 2:
            if offset > self.size:
                offset = self.size
            new_pos = self._filehandle.seek(address + self.size - offset, 0)
        else:
            raise ValueError(
                'invalid whence ({}, should be 0, 1 or 2)'.format(whence)
            )
        self._pos = new_pos - address
        return self._pos

    def tell(self):
        return self._pos

    def __repr__(self):
        return '{}[{}](size={}, only_subobjects={}, address={})'.format(
            self.__class__.__name__,
            self.header.type,
            self.header.content_size,
            self.header.only_subobjects,
            self.header.content_address
        )

    def __str__(self):
        return '{}[{}]'.format(self.__class__.__name__, self.header.type)

    def parse(self):
        if self.only_subobjects:
            raise ValueError('This object only contains subobjects')

        raise NotImplementedError(
            'Parsing of EventIO objects of type {} is not yet implemented'.format(
                self.header.type
            )
        )


def parse_extension_field(extension_field):
    '''parse the so called "extension" field from `file`

    The length of an object can be so large, that it cannot be hold by the
    original `length` field which is 30bits long.
    In that case the most significant part of the length is stored in the
    so called "extension" field. The extension is 12bits long.

    So the total length of the object is:
        real_length = extension * 2^30 + original_lenth

    This function returns the extension *already multiplied with 2^30*
    so that the original length can simply be added to the result of this
    function in order to get the real length of the object.
    '''
    word, = struct.unpack('<I', extension_field)
    extension = get_bits_from_word(
        word, constants.EXTENSION_NUM_BITS, constants.EXTENSION_POS
    )

    # we push the length-extension so many bits to the left,
    # i.e. we multiply with such a high number, that we can simply
    # use the += operator  to combine the normal (small) length and this extension.
    extension <<= constants.LENGTH_NUM_BITS
    return extension



from setuptools import setup, find_packages
import os
import re

# make sure users without cython can install our extensions
try:
    from Cython.Distutils.extension import Extension
    from Cython.Distutils import build_ext as _build_ext
    USE_CYTHON = True
except ImportError:
    from setuptools import Extension
    from setuptools.command.build_ext import build_ext as _build_ext
    USE_CYTHON = False

print('using cython', USE_CYTHON)


# make sure numpy is installed before we try to build
# the extenion
class build_ext(_build_ext):
    def finalize_options(self):
        super().finalize_options()
        import numpy
        self.include_dirs.append(numpy.get_include())


# if we have cython, use the cython file if not the c file
ext = '.pyx' if USE_CYTHON else '.c'
extensions = [
    Extension('eventio.header', sources=['eventio/header' + ext]),
    Extension('eventio.var_int', sources=['eventio/var_int' + ext]),
    Extension(
        'eventio.simtel.parsing',
        sources=['eventio/simtel/parsing' + ext]
    ),
]
cmdclass = {'build_ext': build_ext}

# give a nice error message if people cloned the
# repository and do not have cython installed
if ext == '.c':
    sources = []
    for ext in extensions:
        sources.extend(ext.sources)
    if not all(os.path.isfile(s) for s in sources):
        raise ImportError('You need `Cython` to build this project locally')


with open('README.rst') as f:
    long_description = f.read()


with open('eventio/__init__.py') as f:
    m = re.search("__version__ = [\"']([0-9.]+)[\"']", f.read())
    version = m.groups()[0]

setup(
    name='eventio',
    version=version,
    description='Python read-only implementation of the EventIO file format',
    long_description=long_description,
    url='https://github.com/cta-observatory/pyeventio',
    author='Dominik Neise, Maximilian Noethe',
    author_email='maximilian.noethe@tu-dortmund.de',
    license='MIT',

    packages=find_packages(),

    ext_modules=extensions,
    cmdclass=cmdclass,

    package_data={
        'eventio': ['*.c'],
        'eventio.simtel': ['*.c'],
    },
    python_requires='>=3.5',
    install_requires=[
        'numpy',
        'corsikaio ~= 0.2.0',
    ],
    entry_points={
        'console_scripts': [
            'eventio_print_structure = eventio.scripts.print_structure:main',
            'eventio_plot_histograms = eventio.scripts.plot_hists:main',
            'eventio_print_object_information = eventio.scripts.print_object_information:main',
            'eventio_cut_file = eventio.scripts.cut_eventio_file:main',
        ]
    },
    setup_requires=['pytest-runner', 'numpy'],
    tests_require=['pytest>=3.0.0'],
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3 :: Only',
        'Topic :: Scientific/Engineering :: Astronomy',
        'Topic :: Scientific/Engineering :: Physics',
    ],
)
